from functools import partial
from typing import Any, Callable, Protocol, Tuple, Type, cast

from connexion import ConnexionMiddleware, FlaskApp
from connexion.apps.flask import FlaskASGIApp, FlaskOperation
from connexion.middleware.main import MiddlewarePosition
from flask import Config as Config
from flask import Flask
from flask_injector import FlaskInjector, wrap_function
from injector import Binder, Injector, Module
from Ligare.programming.patterns.dependency_injection import LoggerModule
from starlette.types import ASGIApp, Receive, Scope, Send
from typing_extensions import override

from . import RegisterMiddlewareCallback, TFlaskApp


class MiddlewareRoutine(Protocol):
    def __call__(
        self, scope: Scope, receive: Receive, send: Send, *args: Any
    ) -> None: ...


class AppModule(Module):
    def __init__(self, app: TFlaskApp, *args: Tuple[Any, Any]) -> None:
        super().__init__()
        if isinstance(app, Flask):
            self._flask_app = app
        elif isinstance(app, FlaskApp):  # pyright: ignore[reportUnnecessaryIsInstance] guard against things like not using `MagicMock(spec=...)`
            self._flask_app = app.app
        else:
            raise ValueError(
                f"Wrong type provided for Flask instance. Provided type is `{type(app)}` but excepted `{TFlaskApp}`."
            )

        self._other_dependencies = args

    @override
    def configure(self, binder: Binder) -> None:
        binder.bind(Flask, to=self._flask_app)
        binder.bind(Config, to=self._flask_app.config)
        binder.install(LoggerModule(self._flask_app.name))

        for dependency in self._other_dependencies:
            binder.bind(dependency[0], to=dependency[1])


def configure_dependencies(
    app: TFlaskApp,
    application_modules: list[Module | type[Module]] | None = None,
) -> FlaskInjector:
    """
    Configures dependency injection and registers all Flask
    application dependencies. The FlaskInjector instance
    can be used to bootstrap and start the Flask application.
    """
    if isinstance(app, FlaskApp):
        flask_app = app.app
    else:
        flask_app = app

    modules = [
        (module if isinstance(module, Module) else module())
        for module in [AppModule(app)]
        + (application_modules if application_modules else [])
    ]

    # bootstrap the flask application and its dependencies
    flask_injector = FlaskInjector(flask_app, modules)

    flask_injector.injector.binder.bind(Injector, flask_injector.injector)

    if isinstance(app, FlaskApp):
        app.add_middleware(OpenAPIEndpointDependencyInjectionMiddleware(flask_injector))

        # For every module registered, check if any are "middleware" type modules.
        # if they are, they need to be registered with the application.
        # In the event the application is a Connexion application,
        # this needs to happen after _configure_openapi_middleware_dependencies
        # because the middleware is a "FILO" stack - the items are the end
        # of the stack are executed first. By ensuring this happens before
        # OpenAPIEndpointDependencyInjectionMiddleware, other Middlewares
        # can alter routing information.
        # TODO this needs to happen in some form for plain Flask applications too.
        for module in modules:
            register_callback: RegisterMiddlewareCallback | None = getattr(
                module, "register_middleware", None
            )
            if register_callback is not None and callable(register_callback):
                flask_injector.injector.call_with_injection(
                    register_callback, kwargs={"app": app}
                )

        # this binds all Ligare middlewares with Injector
        _configure_openapi_middleware_dependencies(app, flask_injector)

    return flask_injector


class OpenAPIEndpointDependencyInjectionMiddleware:
    """
    Enables dependency injection for any blueprint methods.
    """

    def __new__(cls, flask_injector: FlaskInjector):
        class _InternalDependencyInjectionMiddleware:
            """
            This is a means of applying partial application over a class, rather than a function.
            It is done this way to jive with Connexion/Starlette's middleware system.
            The inclusion of a flask_injector parameters lets us use that instance
            when setting up the middleware without violating the contract that Starlette expects.
            """

            def __init__(
                self,
                app: ASGIApp,
                flask_injector: FlaskInjector | None = flask_injector,
            ) -> None:
                super().__init__()
                self._app = app
                self._flask_injector = flask_injector

            async def __call__(
                self, scope: Scope, receive: Receive, send: Send
            ) -> None:
                async def wrapped_send(message: Any) -> None:
                    nonlocal scope
                    nonlocal receive
                    nonlocal send

                    # Only run during startup of the application
                    if (
                        (
                            scope["type"] != "lifespan"
                            and message["type"] != "lifespan.startup.complete"
                        )
                        or not scope["app"]
                        or self._flask_injector == None
                    ):
                        return await send(message)

                    # Find every registered route
                    connexion_app = cast(ConnexionMiddleware, scope["app"])

                    if not connexion_app.middleware_stack:
                        return await send(message)

                    flask_asgi_app = next(
                        (
                            middleware
                            for middleware in scope["app"].middleware_stack
                            if isinstance(middleware, FlaskASGIApp)
                        ),
                        None,
                    )

                    if not flask_asgi_app:
                        return await send(message)

                    endpoints = cast(
                        dict[str, FlaskOperation | Callable[..., Any]],
                        flask_asgi_app.app.view_functions,
                    )

                    for endpoint_name, endpoint in endpoints.items():
                        # Skip anything that does not have `@inject` applied
                        if not hasattr(endpoint, "__bindings__"):
                            continue

                        if hasattr(endpoint, "_fn") and isinstance(
                            endpoint, FlaskOperation
                        ):
                            # _fn is the original function that ends up being called.
                            # we wrap it with Injector, then replace it
                            endpoints[endpoint_name]._fn = wrap_function(  # pyright: ignore[reportFunctionMemberAccess]
                                endpoint._fn,
                                self._flask_injector.injector,
                            )
                        # If a blueprint is added through anything other than
                        # by Connexion, `_fn` is not set. This happens with, e.g.,
                        # the SSO blueprint.
                        else:
                            if callable(endpoint):
                                endpoints[endpoint_name] = wrap_function(
                                    endpoint,
                                    self._flask_injector.injector,
                                )

                    return await send(message)

                await self._app(scope, receive, wrapped_send)

        return type(
            "_DependencyInjectionMiddleware",
            (_InternalDependencyInjectionMiddleware,),
            {},
        )


def bind_middleware(
    app: TFlaskApp,
    flask_injector: FlaskInjector,
    middleware: Type[ASGIApp],
    position: MiddlewarePosition = MiddlewarePosition.BEFORE_CONTEXT,
):
    middleware_class: Callable[..., Any] = middleware

    # hasn't been registered with the application yet
    if not isinstance(middleware, partial):
        if not isinstance(app, FlaskApp):
            return

        app.add_middleware(middleware_class, position)
    else:
        middleware_class = middleware.func

    # Connexion/Starlette middleware are classes w/ a __call__ method
    middleware_routine = cast(
        MiddlewareRoutine | None, getattr(middleware_class, "__call__", None)
    )
    if not middleware_routine:
        return

    # Skip any __call__ methods without `@inject` applied
    if not hasattr(middleware_routine, "__bindings__"):
        return

    # Wrap the __call__ method and replace the original with the now-wrapped method
    middleware_class.__call__ = (  # pyright: ignore[reportFunctionMemberAccess]
        wrap_function(middleware_routine, flask_injector.injector)
    )


def _configure_openapi_middleware_dependencies(
    app: TFlaskApp, flask_injector: FlaskInjector
):
    """
    Bind any Connexion middleware classes whose __call__ member has a __bindings__ attribute.
    This attribute signals that @inject was used on it.
    """

    app_middleware = getattr(app, "middleware", None)
    if not app_middleware:
        return

    app_middlewares = getattr(app_middleware, "middlewares", None)
    if not app_middlewares:
        return

    for middleware in cast(list[type], app_middlewares):
        # these are all middlewares that are registered by Connexion
        # so we can skip them
        if not isinstance(middleware, partial):
            continue

        bind_middleware(app, flask_injector, middleware)



"""
This type stub file was generated by pyright.
"""

from connexion.resolver import Resolver

"""
This module contains a mock resolver that returns mock functions for operations it cannot resolve.
"""
logger = ...
class MockResolver(Resolver):
    def __init__(self, mock_all) -> None:
        ...
    
    def resolve(self, operation): # -> Resolution:
        """
        Mock operation resolver

        :type operation: connexion.operations.AbstractOperation
        """
        ...
    
    def mock_operation(self, operation, *args, **kwargs): # -> tuple[Any, Any] | tuple[Literal['No example response was defined.'], Any]:
        ...
    





"""
This type stub file was generated by pyright.
"""

from flask.testing import FlaskClient

class FlaskLoginClient(FlaskClient):
    """
    A Flask test client that knows how to log in users
    using the Flask-Login extension.
    """
    def __init__(self, *args, **kwargs) -> None:
        ...
    





import re
import uuid
from collections.abc import Iterable
from contextlib import ExitStack
from contextvars import Token
from logging import Logger
from typing import Any, Awaitable, Callable, Literal, TypeAlias, TypeVar, cast
from uuid import uuid4

import json_logging
import starlette
import starlette.datastructures
import starlette.requests
from connexion import ConnexionMiddleware, FlaskApp, context, utils
from connexion.middleware import MiddlewarePosition
from flask import Flask, Request, Response, request
from flask.ctx import AppContext
from flask.globals import _cv_app  # pyright: ignore[reportPrivateUsage]
from flask.globals import current_app
from flask.typing import ResponseReturnValue
from flask_login import AnonymousUserMixin, current_user
from injector import inject
from Ligare.programming.collections.dict import AnyDict, merge
from starlette.datastructures import Address
from starlette.types import ASGIApp, Receive, Scope, Send
from typing_extensions import TypedDict, final
from werkzeug.local import LocalProxy

from ...config import Config
from ..consts import (
    CONTENT_SECURITY_POLICY_HEADER,
    CORRELATION_ID_HEADER,
    INCOMING_REQUEST_MESSAGE,
    OUTGOING_RESPONSE_MESSAGE,
    REQUEST_COOKIE_HEADER,
    RESPONSE_COOKIE_HEADER,
)

# pyright: reportUnusedFunction=false


# Fixes type problems when using @inject with @app.before_request and @app.after_request.
# The main difference with these types as opposed to the Flask-defined types is that
# these types allow the handler to take any arguments, versus no arguments or just Response.
AfterRequestCallable: TypeAlias = (
    Callable[..., Response] | Callable[..., Awaitable[Response]]
)
BeforeRequestCallable: TypeAlias = (
    Callable[..., ResponseReturnValue | None]
    | Callable[..., Awaitable[ResponseReturnValue | None]]
)
T_request_callable = TypeVar(
    "T_request_callable", bound=BeforeRequestCallable | AfterRequestCallable | None
)
# Fixes type problems when using Flask-Injector, which automatically sets up any bound
# error handlers. The error handler types in Flask have the same problem as After/BeforeRequestCallable
# in that their parameters are `[Any]` rather than `...`. This relaxes that. The type
# is valid here because Flask-Injector does actually include the provided types as parameters.
ErrorHandlerCallable = (
    Callable[..., ResponseReturnValue] | Callable[..., Awaitable[ResponseReturnValue]]
)
T_error_handler = TypeVar("T_error_handler", bound=ErrorHandlerCallable)

TFlaskApp = Flask | FlaskApp
T_flask_app = TypeVar("T_flask_app", bound=TFlaskApp)


MiddlewareRequestDict = TypedDict(
    "MiddlewareRequestDict",
    {
        "type": Literal["http"],
        "http_version": str,
        "method": Literal[
            "GET",
            "POST",
            "PATCH",
            "PUT",
            "DELETE",
            "OPTIONS",
            "HEAD",
            "CONNECT",
            "TRACE",
        ],
        "path": str,
        "raw_path": bytes,
        "root_path": str,
        "scheme": str,
        "query_string": bytes,
        "headers": list[tuple[bytes, bytes]],
        "client": list[
            str | int
        ],  # this is actually a 2-item list whose first item is a str and second item is an int
        "server": list[str | int],  # same here
        "extensions": dict[str, dict[str, str]],
        "state": AnyDict,
        "app": ConnexionMiddleware,
        "starlette.exception_handlers": tuple[
            dict[type[Any], Callable[..., Any]], dict[Any, Any]
        ],
        "path_params": dict[Any, Any],
    },
)

MiddlewareResponseDict = TypedDict(
    "MiddlewareResponseDict",
    {
        "type": Literal["http.response.start"],
        "status": int,
        "headers": list[tuple[bytes, bytes]],
    },
)


def _get_correlation_id(
    request: MiddlewareRequestDict, response: MiddlewareResponseDict, log: Logger
) -> str:
    correlation_id = _get_correlation_id_from_json_logging(response, log)

    if not correlation_id:
        correlation_id = _get_correlation_id_from_headers(request, response, log)

    return correlation_id


def _get_correlation_id_from_headers(
    request: MiddlewareRequestDict, response: MiddlewareResponseDict, log: Logger
) -> str:
    try:
        headers = _headers_as_dict(request)
        correlation_id = headers.get(CORRELATION_ID_HEADER.lower())

        if not correlation_id:
            headers = _headers_as_dict(response)
            correlation_id = headers.get(CORRELATION_ID_HEADER.lower())

        if correlation_id:
            # validate format
            _ = uuid.UUID(correlation_id)
        else:
            correlation_id = str(uuid4())
            log.info(
                f'Generated new UUID "{correlation_id}" for {CORRELATION_ID_HEADER} request header.'
            )

        return correlation_id

    except ValueError as e:
        log.warning(f"Badly formatted {CORRELATION_ID_HEADER} received in request.")
        raise e


def _get_correlation_id_from_json_logging(
    request_response: MiddlewareRequestDict | MiddlewareResponseDict, log: Logger
) -> str | None:
    correlation_id: None | str
    try:
        correlation_id = json_logging.get_correlation_id(request_response)
        # validate format
        _ = uuid.UUID(correlation_id)
        return correlation_id
    except ValueError as e:
        log.warning(f"Badly formatted {CORRELATION_ID_HEADER} received in request.")
        raise e
    except Exception as e:
        log.debug(
            f"Error received when getting {CORRELATION_ID_HEADER} header from `json_logging`. Possibly `json_logging` is not configured, and this is not an error.",
            exc_info=e,
        )


def _headers_as_dict(
    request_response: MiddlewareRequestDict | MiddlewareResponseDict,
):
    if (
        isinstance(request_response, dict)  # pyright: ignore[reportUnnecessaryIsInstance]
        and "headers" in request_response.keys()
    ):
        # FIXME does this work for a middleware _response_ as well?
        return {
            key: value for (key, value) in decode_headers(request_response["headers"])
        }
    else:
        raise Exception("Unable to extract headers from request when logging request.")


@inject
def _log_all_api_requests(
    request: MiddlewareRequestDict,
    response: MiddlewareResponseDict,
    app: Flask,
    config: Config,
    log: Logger,
):
    request_headers_safe: dict[str, str] = _headers_as_dict(request)

    correlation_id = _get_correlation_id(request, response, log)

    if (
        request_headers_safe.get(REQUEST_COOKIE_HEADER)
        and config.flask
        and config.flask.session
    ):
        request_headers_safe[REQUEST_COOKIE_HEADER] = re.sub(
            rf"({config.flask.session.cookie.name}=)[^;]+(;|$)",
            r"\1<redacted>\2",
            request_headers_safe[REQUEST_COOKIE_HEADER],
        )

    server = get_server_address()
    client = get_remote_address()
    log.info(
        INCOMING_REQUEST_MESSAGE,
        request["method"],
        request["path"],
        # ASGI spec states `server` and `client`
        # can be `None` if not available.
        f"{server.host}:{server.port}",
        f"{client.host}:{client.port}",
        "Anonymous"
        if (
            isinstance(current_user, AnonymousUserMixin)
            or not hasattr(app, "login_manager")
        )
        else current_user.get_id(),
        extra={
            "props": {
                "correlation_id": correlation_id,
                "headers": request_headers_safe,
            }
        },
    )


def _wrap_all_api_responses(
    request: MiddlewareRequestDict,
    response: MiddlewareResponseDict,
    config: Config,
    log: Logger,
):
    correlation_id = _get_correlation_id(request, response, log)
    response_headers = _headers_as_dict(response)

    response_headers[CORRELATION_ID_HEADER] = correlation_id

    if config.web.security.csp:
        response_headers[CONTENT_SECURITY_POLICY_HEADER] = config.web.security.csp

    # if config.flask and config.flask.openapi and config.flask.openapi.use_swagger:
    #    # Use a permissive CSP for the Swagger UI
    #    # https://github.com/swagger-api/swagger-ui/issues/7540
    #    FIXME what to use other than `request` w/ Connexion middleware?
    #    if request.path.startswith("/ui/") or (
    #        request.url_rule and request.url_rule.endpoint == "/v1./v1_swagger_ui_index"
    #    ):
    #        response.headers[
    #            CONTENT_SECURITY_POLICY_HEADER
    #        ] = "default-src 'self'; style-src 'self' 'unsafe-inline'; img-src 'self'; script-src 'self' 'unsafe-inline'"


def _log_all_api_responses(
    request: MiddlewareRequestDict,
    response: MiddlewareResponseDict,
    config: Config,
    log: Logger,
):
    response_headers_safe: dict[str, str] = _headers_as_dict(response)

    correlation_id = _get_correlation_id(request, response, log)

    if (
        response_headers_safe.get(RESPONSE_COOKIE_HEADER)
        and config.flask
        and config.flask.session
    ):
        response_headers_safe[RESPONSE_COOKIE_HEADER] = re.sub(
            rf"({config.flask.session.cookie.name}=)[^;]+(;|$)",
            r"\1<redacted>\2",
            response_headers_safe[RESPONSE_COOKIE_HEADER],
        )

    log.info(
        OUTGOING_RESPONSE_MESSAGE,
        response["status"],
        response["status"],
        extra={
            "props": {
                "correlation_id": correlation_id,
                "headers": response_headers_safe,
            }
        },
    )


def decode_headers(headers: list[tuple[bytes, bytes]]):
    content_type = utils.extract_content_type(headers)
    _, encoding = utils.split_content_type(content_type)
    if encoding is None:
        encoding = "utf-8"

    decoded_headers = [
        (header.decode(encoding), value.decode(encoding)) for (header, value) in headers
    ]

    return decoded_headers


def encode_headers(
    headers: list[tuple[bytes, bytes]],
    append_headers: list[tuple[str, str]],
    encoding: str = "utf-8",
):
    for header, value in append_headers:
        headers.append((header.encode(encoding), value.encode(encoding)))


class RequestLoggerMiddleware:
    _app: ASGIApp

    def __init__(self, app: ASGIApp):
        super().__init__()
        self._app = app

    @inject
    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        config: Config,
        log: Logger,
        app: Flask,
    ) -> None:
        async def wrapped_send(message: Any) -> None:
            nonlocal scope
            nonlocal receive
            nonlocal send

            if message["type"] != "http.response.start":
                return await send(message)

            response = cast(MiddlewareResponseDict, scope)
            request = cast(MiddlewareRequestDict, scope)

            _log_all_api_requests(request, response, app, config, log)

            return await send(message)

        await self._app(scope, receive, wrapped_send)


class ResponseLoggerMiddleware:
    _app: ASGIApp

    def __init__(self, app: ASGIApp):
        super().__init__()
        self._app = app

    @inject
    async def __call__(
        self, scope: Scope, receive: Receive, send: Send, config: Config, log: Logger
    ) -> None:
        async def wrapped_send(message: Any) -> None:
            nonlocal scope
            nonlocal receive
            nonlocal send

            if message["type"] != "http.response.start":
                return await send(message)

            request = cast(MiddlewareRequestDict, scope)
            response = cast(MiddlewareResponseDict, message)

            _log_all_api_responses(request, response, config, log)
            _wrap_all_api_responses(request, response, config, log)

            return await send(message)

        await self._app(scope, receive, wrapped_send)


class CorrelationIDMiddleware:
    _app: ASGIApp

    def __init__(self, app: ASGIApp):
        super().__init__()
        self._app = app

    @inject
    async def __call__(
        self, scope: Scope, receive: Receive, send: Send, log: Logger
    ) -> None:
        async def wrapped_send(message: Any) -> None:
            nonlocal scope
            nonlocal send

            if message["type"] != "http.response.start":
                return await send(message)

            request = cast(MiddlewareRequestDict, scope)
            response = cast(MiddlewareResponseDict, message)

            response_headers = response["headers"]
            content_type = utils.extract_content_type(response_headers)
            _, encoding = utils.split_content_type(content_type)
            if encoding is None:
                encoding = "utf-8"

            request_headers = request["headers"]
            try:
                correlation_id_header_encoded = CORRELATION_ID_HEADER.lower().encode(
                    encoding
                )

                request_correlation_id: bytes | None = next(
                    (
                        correlation_id
                        for (header, correlation_id) in request_headers
                        if header == correlation_id_header_encoded
                    ),
                    None,
                )

                if request_correlation_id:
                    # validate format
                    _ = uuid.UUID(request_correlation_id.decode(encoding))
                else:
                    request_correlation_id = str(uuid4()).encode(encoding)
                    request_headers.append((
                        correlation_id_header_encoded,
                        request_correlation_id,
                    ))
                    log.info(
                        f'Generated new UUID "{request_correlation_id}" for {CORRELATION_ID_HEADER} request header.'
                    )

                response_headers.append((
                    correlation_id_header_encoded,
                    request_correlation_id,
                ))

                return await send(message)
            except ValueError as e:
                log.warning(
                    f"Badly formatted {CORRELATION_ID_HEADER} received in request."
                )
                raise e

        await self._app(scope, receive, wrapped_send)


_DEFAULT_HOSTNAME = "localhost"
_DEFAULT_PORT = 80


def get_server_address() -> Address:
    request_proxy = cast(LocalProxy[starlette.requests.Request], context.request)
    starlette_request = cast(
        starlette.requests.Request, request_proxy._starlette_request
    )
    scope = starlette_request.scope
    server_hostname = starlette_request.base_url.hostname
    server_port = starlette_request.base_url.port
    if not server_hostname or not server_port:
        # there is the possibility this value in scope is "127.0.0.1" when using "localhost"
        # which is not consistent with the _starlette_request value.
        server = cast(tuple[str, int] | None, scope.get("server"))
        if server:
            if not server_hostname:
                server_hostname = server[0]
            if not server_port:
                server_port = server[1]
        else:
            app_server_name = cast(dict[str, str | None], current_app.config).get(
                "SERVER_NAME"
            )
            if app_server_name:
                if not server_hostname:
                    server_hostname = app_server_name[: app_server_name.index(":")]
                if not server_port:
                    server_port = app_server_name[app_server_name.index(":") + 1 :]
    if isinstance(server_port, str):
        try:
            server_port = int(server_port)
        except:
            # this is the same default port that
            # Starlette uses in starlette.middleware.wsgi.build_environ
            server_port = _DEFAULT_PORT

    return Address(
        host=server_hostname or _DEFAULT_HOSTNAME, port=server_port or _DEFAULT_PORT
    )


def get_remote_address() -> Address:
    request_proxy = cast(LocalProxy[starlette.requests.Request], context.request)
    starlette_request = cast(
        starlette.requests.Request, request_proxy._starlette_request
    )
    scope = starlette_request.scope
    request_client = starlette_request.client

    remote_hostname = ""
    remote_port = 0

    if request_client:
        remote_hostname = request_client.host
        remote_port = request_client.port

    if not remote_hostname or not remote_port:
        # it's actually a 2-item list whose first item
        # is a string, and second item is an int ...
        client = cast(tuple[str, int] | None, scope.get("client"))
        if client and isinstance(client, Iterable):  # pyright: ignore[reportUnnecessaryIsInstance]
            if not remote_hostname:
                remote_hostname = client[0]

            if not remote_port:
                remote_port = client[1]

    return Address(host=remote_hostname, port=remote_port)


def address_to_str(address: Address):
    return f"{address.host}:{address.port}"


@final
class FlaskContextMiddleware:
    """
    Connexion does not set Flask contexts in all cases they may be needed, like
    in middlewares that execute before ContextMiddleware. This middleware creates
    the Flask application, request, and session contexts if they are not currently set.
    """

    def __init__(
        self,
        app: ASGIApp,
    ) -> None:
        self.app = app

    @inject
    async def __call__(
        self, scope: Scope, receive: Receive, send: Send, app: Flask
    ) -> None:
        receive_token: Token[Receive] | None = None
        scope_token: Token[Scope] | None = None
        app_ctx_token: Token[AppContext] | None = None

        try:
            receive_token = context._receive.set(receive)  # pyright: ignore[reportPrivateUsage]
            scope_token = context._scope.set(scope)  # pyright: ignore[reportPrivateUsage]

            if scope["type"] not in ["http", "websocket"]:
                await self.app(scope, receive, send)
                return

            with ExitStack() as exit_stack:
                if not isinstance(current_app, Flask):  # pyright: ignore[reportUnnecessaryIsInstance] it is not a Flask if it is not set
                    app_ctx = exit_stack.enter_context(app.app_context())

                    app_ctx_token = _cv_app.set(app_ctx)

                if not isinstance(request, Request):  # pyright: ignore[reportUnnecessaryIsInstance] it is not a Request if it is not set
                    request_proxy = cast(
                        LocalProxy[starlette.requests.Request], context.request
                    )
                    starlette_request = cast(
                        starlette.requests.Request, request_proxy._starlette_request
                    )
                    request_headers = cast(
                        starlette.datastructures.Headers, request_proxy.headers
                    )
                    path_info = (
                        "PATH_INFO",
                        str(scope.get("path") or starlette_request.url.path),
                    )
                    wsgi_url_scheme = (
                        "wsgi.url_scheme",
                        str(scope.get("scheme") or starlette_request.url.scheme),
                    )
                    request_method = (
                        "REQUEST_METHOD",
                        str(scope.get("method") or starlette_request.method),
                    )
                    (server_hostname, server_port) = get_server_address()
                    server_name = ("SERVER_NAME", server_hostname)
                    server_port = ("SERVER_PORT", server_port)
                    query_string = (
                        "QUERY_STRING",
                        scope.get("query_string") or starlette_request.url.query,
                    )
                    remote_addr = ("REMOTE_ADDR", address_to_str(get_remote_address()))
                    headers = dict({
                        (
                            f"{'HTTP_' if header.upper() not in ['CONTENT_TYPE', 'CONTENT_LENGTH'] else ''}{header.upper().replace('-', '_')}",
                            value,
                        )
                        for header, value in request_headers.items()
                    })

                    request_environ = merge(
                        dict([
                            path_info,
                            wsgi_url_scheme,
                            request_method,
                            server_name,
                            server_port,
                            query_string,
                            remote_addr,
                        ]),
                        headers,
                        True,
                    )

                    # Some values, like the query string, are stored as bytes.
                    # Decode as a UTF-8 str so encoding later doesn't fail.
                    for key, value in request_environ.items():
                        if isinstance(value, bytes):
                            request_environ[key] = value.decode("utf-8")

                    request_ctx = exit_stack.enter_context(
                        app.request_context(request_environ)
                    )
                    request_ctx.push()

                await self.app(scope, receive, send)

        finally:
            if app_ctx_token is not None:
                _cv_app.reset(app_ctx_token)
            if receive_token is not None:
                context._receive.reset(receive_token)  # pyright: ignore[reportPrivateUsage]
            if scope_token is not None:
                context._scope.reset(scope_token)  # pyright: ignore[reportPrivateUsage]


def register_openapi_api_request_handlers(app: FlaskApp):
    app.add_middleware(RequestLoggerMiddleware)


def register_openapi_api_response_handlers(app: FlaskApp):
    app.add_middleware(CorrelationIDMiddleware)
    app.add_middleware(ResponseLoggerMiddleware)


def register_openapi_context_middleware(app: FlaskApp):
    app.add_middleware(FlaskContextMiddleware, MiddlewarePosition.BEFORE_EXCEPTION)



import logging
from collections import defaultdict
from dataclasses import dataclass
from os import environ, path
from typing import (
    Any,
    Generator,
    Generic,
    Optional,
    Protocol,
    TypeVar,
    cast,
    final,
    overload,
)

import json_logging
from connexion import FlaskApp
from flask import Blueprint, Flask
from flask_injector import FlaskInjector
from injector import Module
from lib_programname import get_path_executed_script
from Ligare.AWS.ssm import SSMParameters
from Ligare.programming.collections.dict import NestedDict
from Ligare.programming.config import (
    AbstractConfig,
    ConfigBuilder,
    TConfig,
    load_config,
)
from Ligare.programming.config.exceptions import ConfigBuilderStateError
from Ligare.programming.dependency_injection import ConfigModule
from Ligare.programming.patterns.dependency_injection import ConfigurableModule
from Ligare.web.exception import BuilderBuildError, InvalidBuilderStateError
from typing_extensions import Self, deprecated

from .config import Config
from .middleware import (
    register_api_request_handlers,
    register_api_response_handlers,
    register_context_middleware,
    register_error_handlers,
)
from .middleware.dependency_injection import configure_dependencies

_get_program_dir = lambda: path.dirname(get_path_executed_script())
_get_exec_dir = lambda: path.abspath(".")

TApp = Flask | FlaskApp
T_app = TypeVar("T_app", bound=TApp)

TAppConfig = TypeVar("TAppConfig", bound=Config)


@dataclass
class AppInjector(Generic[T_app]):
    """
    Contains an instantiated `T_app` application in `app`,
    and its associated `FlaskInjector` IoC container.

    :param T_app Generic: An instance of Flask or FlaskApp.
    :param flask_inject FlaskInject: The applications IoC container.
    """

    app: T_app
    flask_injector: FlaskInjector


@dataclass
class CreateAppResult(Generic[T_app]):
    """
    Contains an instantiated Flask application and its
    associated application "container." This is either
    the same Flask instance, or an OpenAPI application.

    :param flask_app Generic: The Flask application.
    :param app_injector AppInjector[T_app]: The application's wrapper and IoC container.
    """

    flask_app: Flask
    app_injector: AppInjector[T_app]


FlaskAppResult = CreateAppResult[Flask]
OpenAPIAppResult = CreateAppResult[FlaskApp]


# In Python 3.12 we can use generics in functions,
# but we target >= Python 3.10. This is one way
# around that limitation.
@deprecated("`App` is deprecated. Use `ApplicationBuilder`.")
class App(Generic[T_app]):
    """
    Create a new generic type for the application instance.

    Type Args:
        T_app: Either `Flask` or `FlaskApp`
    """

    @deprecated("`App.create` is deprecated. Use `ApplicationBuilder`.")
    @staticmethod
    def create(
        config_filename: str = "config.toml",
        # FIXME should be a list of PydanticDataclass
        application_configs: list[type[AbstractConfig]] | None = None,
        application_modules: list[Module | type[Module]] | None = None,
    ) -> CreateAppResult[T_app]:
        """
        Bootstrap the Flask applcation.

        Args:
            config_filename: The name of the TOML file to load configuration information from.
            application_config: A list of Pydantic objects to store configuration information from the TOML file.
            application_modules: Modules the application will use for the application lifetime.
        """
        return cast(
            CreateAppResult[T_app],
            _create_app(config_filename, application_configs, application_modules),
        )


class UseConfigurationCallback(Protocol[TConfig]):
    """
    The callback for configuring an application's configuration.

    :param TConfig Protocol: The AbstractConfig type to be configured.
    """

    def __call__(
        self,
        config_builder: ConfigBuilder[TConfig],
        config_overrides: dict[str, Any],
    ) -> "None | ConfigBuilder[TConfig]":
        """
        Set up parameters for the application's configuration.

        :param ConfigBuilder[TConfig] config_builder: The ConfigBuilder instance.
        :param dict[str, Any] config_overrides: A dictionary of key/values that are applied over all keys that might exist in an instantiated config.
        :raises InvalidBuilderStateError: Upon a call to `build()`, the builder is misconfigured.
        :raises BuilderBuildError: Upon a call to `build()`, a failure occurred during the instantiation of the configuration.
        :raises Exception: Upon a call to `build()`, an unknown error occurred.
        :return None | ConfigBuilder[TConfig]: The callback may return `None` or the received `ConfigBuilder` instance so as to support the use of lambdas. This return value is not used.
        """


@final
class ApplicationConfigBuilder(Generic[TConfig]):
    _DEFAULT_CONFIG_FILENAME: str = "config.toml"

    def __init__(self) -> None:
        self._config_value_overrides: dict[str, Any] = {}
        self._config_builder: ConfigBuilder[TConfig] = ConfigBuilder[TConfig]()
        self._config_filename: str = ApplicationConfigBuilder._DEFAULT_CONFIG_FILENAME
        self._use_filename: bool = False
        self._use_ssm: bool = False

    def with_config_builder(self, config_builder: ConfigBuilder[TConfig]) -> Self:
        self._config_builder = config_builder
        return self

    def with_root_config_type(self, config_type: type[TConfig]) -> Self:
        _ = self._config_builder.with_root_config(config_type)
        return self

    def with_config_types(self, configs: list[type[AbstractConfig]] | None) -> Self:
        _ = self._config_builder.with_configs(configs)
        return self

    def with_config_type(self, config_type: type[AbstractConfig]) -> Self:
        _ = self._config_builder.with_config(config_type)
        return self

    def with_config_value_overrides(self, values: dict[str, Any]) -> Self:
        self._config_value_overrides = values
        return self

    def with_config_filename(self, filename: str) -> Self:
        self._config_filename = filename
        self._use_filename = True
        return self

    def enable_ssm(self, value: bool) -> Self:
        """
        Try to load config from AWS SSM. If `use_filename` was configured,
        a failed attempt to load from SSM will instead attempt to load from
        the configured filename. If `use_filename` is not configured and SSM
        fails, an exception is raised. If SSM succeeds, `build` will not
        load from the configured filename.

        :param bool value: Whether to use SSM
        :return Self:
        """
        self._use_ssm = value
        return self

    def build(self) -> TConfig | None:
        if not (self._use_ssm or self._use_filename):
            raise InvalidBuilderStateError(
                "Cannot build the application config without either `use_ssm` or `use_filename` having been configured."
            )

        try:
            config_type = self._config_builder.build()
        except ConfigBuilderStateError as e:
            raise BuilderBuildError(
                "A root config must be specified using `with_root_config` before calling `build()`."
            ) from e

        full_config: TConfig | None = None
        SSM_FAIL_ERROR_MSG = "Unable to load configuration. SSM parameter load failed and the builder is configured not to load from a file."
        if self._use_ssm:
            try:
                # requires that aws-ssm.ini exists and is correctly configured
                ssm_parameters = SSMParameters()
                full_config = ssm_parameters.load_config(config_type)

                if not self._use_filename and full_config is None:
                    raise BuilderBuildError(SSM_FAIL_ERROR_MSG)
            except Exception as e:
                if self._use_filename:
                    logging.getLogger().info("SSM parameter load failed.", exc_info=e)
                else:
                    raise BuilderBuildError(SSM_FAIL_ERROR_MSG) from e

        if self._use_filename and full_config is None:
            if self._config_value_overrides:
                full_config = load_config(
                    config_type, self._config_filename, self._config_value_overrides
                )
            else:
                full_config = load_config(config_type, self._config_filename)

        return full_config


class ApplicationConfigBuilderCallback(Protocol[TAppConfig]):
    def __call__(
        self,
        config_builder: ApplicationConfigBuilder[TAppConfig],
    ) -> "None | ApplicationConfigBuilder[TAppConfig]": ...


@final
class ApplicationBuilder(Generic[T_app]):
    def __init__(self) -> None:
        self._modules: list[Module | type[Module]] = []
        self._config_overrides: dict[str, Any] = {}

    _APPLICATION_CONFIG_BUILDER_PROPERTY_NAME: str = "__application_config_builder"

    @property
    def _application_config_builder(self) -> ApplicationConfigBuilder[Config]:
        builder = getattr(
            self, ApplicationBuilder._APPLICATION_CONFIG_BUILDER_PROPERTY_NAME, None
        )

        if builder is None:
            builder = ApplicationConfigBuilder[Config]()
            self._application_config_builder = builder.with_root_config_type(Config)

        return builder

    @_application_config_builder.setter
    def _application_config_builder(self, value: ApplicationConfigBuilder[Config]):
        setattr(
            self, ApplicationBuilder._APPLICATION_CONFIG_BUILDER_PROPERTY_NAME, value
        )

    @overload
    def with_module(self, module: Module) -> Self: ...
    @overload
    def with_module(self, module: type[Module]) -> Self: ...
    def with_module(self, module: Module | type[Module]) -> Self:
        module_type = type(module) if isinstance(module, Module) else module
        if issubclass(module_type, ConfigurableModule):
            _ = self._application_config_builder.with_config_type(
                module_type.get_config_type()
            )

        self._modules.append(module)
        return self

    def with_modules(self, modules: list[Module | type[Module]] | None) -> Self:
        if modules is not None:
            for module in modules:
                _ = self.with_module(module)
        return self

    @overload
    def use_configuration(
        self,
        __application_config_builder_callback: ApplicationConfigBuilderCallback[Config],
    ) -> Self:
        """
        Execute changes to the builder's `ApplicationConfigBuilder[TAppConfig]` instance.

        `__builder_callback` can return `None`, or the instance of `ApplicationConfigBuilder[TAppConfig]` passed to its `config_builder` argument.
        This allowance is so lambdas can be used; `ApplicationBuilder[T_app, TAppConfig]` does not use the return value.
        """
        ...

    @overload
    def use_configuration(
        self, __application_config_builder: ApplicationConfigBuilder[Config]
    ) -> Self:
        """Replace the builder's default `ApplicationConfigBuilder[TAppConfig]` instance, or any instance previously assigned."""
        ...

    def use_configuration(
        self,
        application_config_builder: ApplicationConfigBuilderCallback[Config]
        | ApplicationConfigBuilder[Config],
    ) -> Self:
        if callable(application_config_builder):
            _ = application_config_builder(self._application_config_builder)
        else:
            self._application_config_builder = application_config_builder

        return self

    def with_flask_app_name(self, value: str | None) -> Self:
        self._config_overrides["app_name"] = value
        return self

    def with_flask_env(self, value: str | None) -> Self:
        self._config_overrides["env"] = value
        return self

    def build(self) -> CreateAppResult[T_app]:
        config_overrides: NestedDict[str, Any] = defaultdict(dict)

        if (
            override_app_name := self._config_overrides.get("app_name", None)
        ) is not None and override_app_name != "":
            config_overrides["flask"]["app_name"] = override_app_name

        if (
            override_env := self._config_overrides.get("env", None)
        ) is not None and override_env != "":
            config_overrides["flask"]["env"] = override_env

        _ = self._application_config_builder.with_config_value_overrides(
            config_overrides
        )
        config = self._application_config_builder.build()

        if config is None:
            raise BuilderBuildError(
                "Failed to load the application configuration correctly."
            )

        if config.flask is None:
            raise Exception("You must set [flask] in the application configuration.")

        if not config.flask.app_name:
            raise Exception(
                "You must set the Flask application name in the [flask.app_name] config or FLASK_APP envvar."
            )

        app: T_app

        if config.flask.openapi is not None:
            openapi = configure_openapi(config)
            app = cast(T_app, openapi)
        else:
            app = cast(T_app, configure_blueprint_routes(config))

        register_error_handlers(app)
        _ = register_api_request_handlers(app)
        _ = register_api_response_handlers(app)
        _ = register_context_middleware(app)

        application_modules = [
            ConfigModule(config, type(config))
            for (_, config) in cast(
                Generator[tuple[str, AbstractConfig], None, None], config
            )
        ] + (self._modules if self._modules else [])
        # The `config` module cannot be overridden unless the application
        # IoC container is fiddled with. `config` is the instance registered
        # to `AbstractConfig`.
        modules = application_modules + [ConfigModule(config, Config)]
        flask_injector = configure_dependencies(app, application_modules=modules)

        flask_app = app.app if isinstance(app, FlaskApp) else app
        return CreateAppResult[T_app](
            flask_app, AppInjector[T_app](app, flask_injector)
        )


@deprecated("`create_app` is deprecated. Use `ApplicationBuilder`.")
def create_app(
    config_filename: str = "config.toml",
    # FIXME should be a list of PydanticDataclass
    application_configs: list[type[AbstractConfig]] | None = None,
    application_modules: list[Module | type[Module]] | None = None,
) -> CreateAppResult[TApp]:
    """
    Do not use this method directly. Instead, use `App[T_app].create()` or `ApplicationBuilder[TApp, TConfig]()`
    """
    return _create_app(config_filename, application_configs, application_modules)


def _create_app(
    config_filename: str = "config.toml",
    # FIXME should be a list of PydanticDataclass
    application_configs: list[type[AbstractConfig]] | None = None,
    application_modules: list[Module | type[Module]] | None = None,
) -> CreateAppResult[TApp]:
    # set up the default configuration as soon as possible
    # also required to call before json_logging.config_root_logger()
    logging.basicConfig(force=True)

    application_builder = (
        ApplicationBuilder[TApp]()
        .with_flask_app_name(environ.get("FLASK_APP", None))
        .with_flask_env(environ.get("FLASK_ENV", None))
        .with_modules(application_modules)
        .use_configuration(
            lambda config_builder: config_builder.enable_ssm(True)
            .with_config_filename(config_filename)
            .with_root_config_type(Config)
            .with_config_types(application_configs)
        )
    )
    app = application_builder.build()
    return app


def configure_openapi(config: Config, name: Optional[str] = None):
    """
    Instantiate Connexion and set Flask logging options
    """

    if (
        config.flask is None
        or config.flask.openapi is None
        or config.flask.openapi.spec_path is None
    ):
        raise Exception(
            "OpenAPI configuration is empty. Review the `openapi` section of your application's `config.toml`."
        )

    exec_dir = _get_exec_dir()

    connexion_app = FlaskApp(
        config.flask.app_name,
        specification_dir=exec_dir,
        # host=host,
        # port=port,
    )
    app = connexion_app.app
    config.update_flask_config(app.config)

    enable_json_logging = config.logging.format == "JSON"
    if enable_json_logging:
        json_logging.init_connexion(  # pyright: ignore[reportUnknownMemberType]
            enable_json=enable_json_logging
        )
        json_logging.init_request_instrument(  # pyright: ignore[reportUnknownMemberType]
            connexion_app
        )
        json_logging.config_root_logger()  # pyright: ignore[reportUnknownMemberType]

    app.logger.setLevel(environ.get("LOGLEVEL", "INFO").upper())

    options: dict[str, bool | str] = {
        "swagger_ui": config.flask.openapi.use_swagger,
        "swagger_url": config.flask.openapi.swagger_url or "/",
    }

    _ = connexion_app.add_api(
        f"{config.flask.app_name}/{config.flask.openapi.spec_path}",
        # base_path="/",
        validate_responses=config.flask.openapi.validate_responses,
        options=options,
    )

    # FIXME what's the new way to get this URL?
    # if config.flask.openapi.use_swagger:
    #    # App context needed for url_for.
    #    # This can only run after connexion is instantiated
    #    # because it registers the swagger UI url.
    #    with app.app_context():
    #        app.logger.info(
    #            f"Swagger UI can be accessed at {url_for('/./_swagger_ui_index', _external=True)}"
    #        )

    return connexion_app


def configure_blueprint_routes(
    config: Config, blueprint_import_subdir: str = "endpoints"
):
    """
    Register Flask blueprints and API routes
    """
    if config.flask is None:
        raise Exception(
            "Flask configuration is empty. Review the `flask` section of your application's `config.toml`."
        )

    app = Flask(config.flask.app_name)
    config.update_flask_config(app.config)

    enable_json_logging = config.logging.format == "JSON"
    if enable_json_logging:
        json_logging.init_flask(  # pyright: ignore[reportUnknownMemberType]
            enable_json=enable_json_logging
        )
        json_logging.init_request_instrument(  # pyright: ignore[reportUnknownMemberType]
            app
        )
        json_logging.config_root_logger()  # pyright: ignore[reportUnknownMemberType]

    blueprint_modules = _import_blueprint_modules(app, blueprint_import_subdir)
    _register_blueprint_modules(app, blueprint_modules)
    return app


def _import_blueprint_modules(app: Flask, blueprint_import_subdir: str):
    from importlib.util import module_from_spec, spec_from_file_location
    from pathlib import Path

    from flask import Blueprint

    program_dir = _get_program_dir()
    blueprint_import_dir = Path(program_dir, blueprint_import_subdir)

    module_paths = blueprint_import_dir.glob("*.py")

    blueprint_modules: list[Blueprint] = []

    for path in module_paths:
        if not (path.is_file() or path.name == "__init__.py"):
            continue
        # load the module from its path
        # and execute it
        spec = spec_from_file_location(path.name.rstrip(".py"), str(path))
        if spec is None or spec.loader is None:
            raise Exception(f"Module cannot be created from path {path}")
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        # find all Flask blueprints in
        # the module and register them
        for module_name, module_var in vars(module).items():
            if module_name.endswith("_blueprint") or isinstance(module_var, Blueprint):
                blueprint_modules.append(module_var)

    return blueprint_modules


def _register_blueprint_modules(app: Flask, blueprint_modules: list[Blueprint]):
    for module in blueprint_modules:
        app.register_blueprint(module)



"""
This type stub file was generated by pyright.
"""

import json
import logging
import sys
import traceback
import uuid
from datetime import datetime

import json_logging.framework.connexion as connexion_support
import json_logging.framework.fastapi as fastapi_support
import json_logging.framework.flask as flask_support
import json_logging.framework.quart as quart_support
from json_logging import util
from json_logging.framework.sanic import (
    SanicAppConfigurator,
    SanicAppRequestInstrumentationConfigurator,
    SanicRequestAdapter,
    SanicResponseAdapter,
)
from json_logging.framework_base import (
    AppRequestInstrumentationConfigurator,
    FrameworkConfigurator,
    RequestAdapter,
    ResponseAdapter,
)
from json_logging.util import get_library_logger, is_env_var_toggle

CORRELATION_ID_GENERATOR = ...
ENABLE_JSON_LOGGING = ...
if is_env_var_toggle("ENABLE_JSON_LOGGING"):
    ENABLE_JSON_LOGGING = ...
ENABLE_JSON_LOGGING_DEBUG = ...
EMPTY_VALUE = ...
CREATE_CORRELATION_ID_IF_NOT_EXISTS = ...
JSON_SERIALIZER = ...
CORRELATION_ID_HEADERS = ...
COMPONENT_ID = ...
COMPONENT_NAME = ...
COMPONENT_INSTANCE_INDEX = ...
_framework_support_map = ...
_current_framework = ...
_logger = ...
_request_util = ...
_default_formatter = ...

def get_correlation_id(request=...) -> str:
    """
    Get current request correlation-id. If one is not present, a new one might be generated
    depends on CREATE_CORRELATION_ID_IF_NOT_EXISTS setting value.

    :return: correlation-id string
    """
    ...

def register_framework_support(
    name,
    app_configurator,
    app_request_instrumentation_configurator,
    request_adapter_class,
    response_adapter_class,
):  # -> None:
    """
    register support for a framework

    :param name: name of framework
    :param app_configurator: app pre-configurator class
    :param app_request_instrumentation_configurator: app configurator class
    :param request_adapter_class: request adapter class
    :param response_adapter_class: response adapter class
    """
    ...

def config_root_logger():  # -> None:
    """
    You must call this if you are using root logger.
    Make all root logger' handlers produce JSON format
    & remove duplicate handlers for request instrumentation logging.
    Please made sure that you call this after you called "logging.basicConfig() or logging.getLogger()
    """
    ...

def init_non_web(*args, **kw):  # -> None:
    ...

def init_request_instrument(
    app=..., custom_formatter=..., exclude_url_patterns=...
):  # -> None:
    """
    Configure the request instrumentation logging configuration for given web app. Must be called after init method

    If **custom_formatter** is passed, it will use this formatter over the default.

    :param app: current web application instance
    :param custom_formatter: formatter to override default JSONRequestLogFormatter.
    """
    ...

def get_request_logger(): ...

class RequestInfo(dict):
    """
    class that keep HTTP request information for request instrumentation logging
    """

    def __init__(self, request, **kwargs) -> None: ...
    def update_response_status(self, response):  # -> None:
        """
        update response information into this object, must be called before invoke request logging statement
        :param response:
        """
        ...

class BaseJSONFormatter(logging.Formatter):
    """
    Base class for JSON formatters
    """

    base_object_common = ...
    def __init__(self, *args, **kw) -> None: ...
    def format(self, record):  # -> str:
        ...

class JSONRequestLogFormatter(BaseJSONFormatter):
    """
    Formatter for HTTP request instrumentation logging
    """

    ...

class JSONLogFormatter(BaseJSONFormatter):
    """
    Formatter for non-web application log
    """

    def get_exc_fields(self, record):  # -> dict[str, Unknown]:
        ...
    @classmethod
    def format_exception(cls, exc_info):  # -> str:
        ...

class JSONLogWebFormatter(JSONLogFormatter):
    """
    Formatter for web application log
    """

    ...

def init_flask(custom_formatter=..., enable_json=...):  # -> None:
    ...

def init_sanic(custom_formatter=..., enable_json=...):  # -> None:
    ...

def init_quart(custom_formatter=..., enable_json=...):  # -> None:
    ...

def init_connexion(custom_formatter=..., enable_json=...):  # -> None:
    ...

if fastapi_support.is_fastapi_present(): ...

def init_fastapi(custom_formatter=..., enable_json=...):  # -> None:
    ...



"""
This type stub file was generated by pyright.
"""

import typing as t
from contextvars import ContextVar
from starlette.types import ASGIApp, Receive, Scope, Send
from connexion.middleware.abstract import AbstractRoutingAPI, SpecMiddleware
from connexion.operations import AbstractOperation
from connexion.resolver import Resolver
from connexion.spec import Specification

_scope: ContextVar[dict] = ...
class RoutingOperation:
    def __init__(self, operation_id: t.Optional[str], next_app: ASGIApp) -> None:
        ...
    
    @classmethod
    def from_operation(cls, operation: AbstractOperation, next_app: ASGIApp): # -> Self:
        ...
    
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Attach operation to scope and pass it to the next app"""
        ...
    


class RoutingAPI(AbstractRoutingAPI):
    def __init__(self, specification: Specification, *, next_app: ASGIApp, base_path: t.Optional[str] = ..., arguments: t.Optional[dict] = ..., resolver: t.Optional[Resolver] = ..., resolver_error_handler: t.Optional[t.Callable] = ..., debug: bool = ..., **kwargs) -> None:
        """API implementation on top of Starlette Router for Connexion middleware."""
        ...
    
    def make_operation(self, operation: AbstractOperation) -> RoutingOperation:
        ...
    


class RoutingMiddleware(SpecMiddleware):
    def __init__(self, app: ASGIApp) -> None:
        """Middleware that resolves the Operation for an incoming request and attaches it to the
        scope.

        :param app: app to wrap in middleware.
        """
        ...
    
    def add_api(self, specification: Specification, base_path: t.Optional[str] = ..., arguments: t.Optional[dict] = ..., **kwargs) -> None:
        """Add an API to the router based on a OpenAPI spec.

        :param specification: OpenAPI spec.
        :param base_path: Base path where to add this API.
        :param arguments: Jinja arguments to replace in the spec.
        """
        ...
    
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Route request to matching operation, and attach it to the scope before calling the
        next app."""
        ...
    





import json
from logging import Logger
from typing import Awaitable, Callable, TypeAlias, TypeVar

from connexion import FlaskApp
from flask import Flask, Response
from flask.typing import ResponseReturnValue
from Ligare.web.middleware.flask import (
    register_flask_api_request_handlers,
    register_flask_api_response_handlers,
)
from Ligare.web.middleware.openapi import (
    register_openapi_api_request_handlers,
    register_openapi_api_response_handlers,
    register_openapi_context_middleware,
)
from werkzeug.exceptions import HTTPException, Unauthorized

# pyright: reportUnusedFunction=false


# Fixes type problems when using @inject with @app.before_request and @app.after_request.
# The main difference with these types as opposed to the Flask-defined types is that
# these types allow the handler to take any arguments, versus no arguments or just Response.
AfterRequestCallable: TypeAlias = (
    Callable[..., Response] | Callable[..., Awaitable[Response]]
)
BeforeRequestCallable: TypeAlias = (
    Callable[..., ResponseReturnValue | None]
    | Callable[..., Awaitable[ResponseReturnValue | None]]
)
T_request_callable = TypeVar(
    "T_request_callable", bound=BeforeRequestCallable | AfterRequestCallable | None
)
# Fixes type problems when using Flask-Injector, which automatically sets up any bound
# error handlers. The error handler types in Flask have the same problem as After/BeforeRequestCallable
# in that their parameters are `[Any]` rather than `...`. This relaxes that. The type
# is valid here because Flask-Injector does actually include the provided types as parameters.
ErrorHandlerCallable = (
    Callable[..., ResponseReturnValue] | Callable[..., Awaitable[ResponseReturnValue]]
)
T_error_handler = TypeVar("T_error_handler", bound=ErrorHandlerCallable)

TFlaskApp = Flask | FlaskApp
T_flask_app = TypeVar("T_flask_app", bound=TFlaskApp)

RegisterMiddlewareCallback = Callable[[FlaskApp], None]


def bind_errorhandler(
    app: TFlaskApp,
    code_or_exception: type[Exception] | int,
) -> Callable[[T_error_handler], T_error_handler | None]:
    if isinstance(app, Flask):
        return app.errorhandler(code_or_exception)
    else:
        return app.app.errorhandler(code_or_exception)


def register_api_request_handlers(app: TFlaskApp):
    if isinstance(app, Flask):
        return register_flask_api_request_handlers(app)
    else:
        return register_openapi_api_request_handlers(app)


def register_api_response_handlers(app: TFlaskApp):
    if isinstance(app, Flask):
        # TODO consider moving request/response logging to the WSGI app
        # apparently Flask may not call this if unhandled exceptions occur
        return register_flask_api_response_handlers(app)
    else:
        return register_openapi_api_response_handlers(app)


def register_context_middleware(app: TFlaskApp):
    if not isinstance(app, FlaskApp):
        return
    return register_openapi_context_middleware(app)


def register_error_handlers(app: TFlaskApp):
    @bind_errorhandler(app, Exception)
    # @inject
    def catch_all_catastrophic(error: Exception, log: Logger):
        # error: connexion.lifecycle.ConnexionRequest, log: ZeroDivisionError
        log.exception(error)

        response = {
            "status_code": 500,
            "error_msg": "Unknown error.",
            "status": "Internal Server Error",
        }
        return response, 500

    @bind_errorhandler(app, HTTPException)
    # @inject
    def catch_all(error: HTTPException, log: Logger):
        log.exception(error)

        response = {
            "status_code": error.code,
            "error_msg": error.description,
            "status": error.name,
        }

        return (
            response,
            error.code if error.code is not None else 500,
        )

    @bind_errorhandler(app, 401)
    # @inject
    def unauthorized(error: Unauthorized, log: Logger):
        log.info(error)

        if error.response is None or not isinstance(error.response, Response):
            response = {
                "status_code": 401,
                "error_msg": error.description,
                "status": error.name,
            }
            return response, 401

        response = error.response
        data = {
            "status_code": response.status_code,
            "error_msg": response.data.decode(),
            "status": response.status,
        }
        response.data = json.dumps(data)
        return response



"""
This type stub file was generated by pyright.
"""

import typing as t
from starlette.exceptions import HTTPException
from starlette.middleware.exceptions import ExceptionMiddleware as StarletteExceptionMiddleware
from starlette.requests import Request as StarletteRequest
from starlette.responses import Response as StarletteResponse
from starlette.types import ASGIApp, Receive, Scope, Send
from connexion.exceptions import ProblemException
from connexion.lifecycle import ConnexionRequest, ConnexionResponse
from connexion.types import MaybeAwaitable

logger = ...
def connexion_wrapper(handler: t.Callable[[ConnexionRequest, Exception], MaybeAwaitable[ConnexionResponse]]) -> t.Callable[[StarletteRequest, Exception], t.Awaitable[StarletteResponse]]:
    """Wrapper that translates Starlette requests to Connexion requests before passing
    them to the error handler, and translates the returned Connexion responses to
    Starlette responses."""
    ...

class ExceptionMiddleware(StarletteExceptionMiddleware):
    """Subclass of starlette ExceptionMiddleware to change handling of HTTP exceptions to
    existing connexion behavior."""
    def __init__(self, next_app: ASGIApp) -> None:
        ...
    
    def add_exception_handler(self, exc_class_or_status_code: t.Union[int, t.Type[Exception]], handler: t.Callable[[ConnexionRequest, Exception], StarletteResponse]) -> None:
        ...
    
    @staticmethod
    def problem_handler(_request: ConnexionRequest, exc: ProblemException): # -> ConnexionResponse:
        """Default handler for Connexion ProblemExceptions"""
        ...
    
    @staticmethod
    @connexion_wrapper
    def http_exception(_request: StarletteRequest, exc: HTTPException, **kwargs) -> StarletteResponse:
        """Default handler for Starlette HTTPException"""
        ...
    
    @staticmethod
    def common_error_handler(_request: StarletteRequest, exc: Exception) -> ConnexionResponse:
        """Default handler for any unhandled Exception"""
        ...
    
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        ...
    





class InvalidBuilderStateError(Exception):
    """The builder's state is invalid and the builder cannot execute `build()`."""


class BuilderBuildError(Exception):
    """The builder failed during execution of `build()`."""



from flask import Flask
from injector import Injector
from Ligare.web.middleware.dependency_injection import AppModule
from mock import MagicMock


def test__AppModule__binds_extra_dependencies():
    flask_mock = MagicMock(spec=Flask)
    flask_mock.name = f"{test__AppModule__binds_extra_dependencies.__name__}-app_name"
    flask_mock.config = {}
    extra_dependency_mock = MagicMock()

    class TestDependencyType: ...

    app_module = AppModule(flask_mock, (TestDependencyType, extra_dependency_mock))

    injector = Injector(app_module)
    resolved_type_instance = injector.get(TestDependencyType)

    assert id(resolved_type_instance) == id(extra_dependency_mock)



"""
This type stub file was generated by pyright.
"""

import abc
import typing as t
from connexion.frameworks.abstract import Framework
from connexion.lifecycle import ConnexionRequest, WSGIRequest
from connexion.operations import AbstractOperation

"""
This module defines a decorator to convert request parameters to arguments for the view function.
"""
logger = ...
CONTEXT_NAME = ...
class BaseParameterDecorator:
    def __init__(self, *, framework: t.Type[Framework], pythonic_params: bool = ...) -> None:
        ...
    
    @abc.abstractmethod
    def __call__(self, function: t.Callable) -> t.Callable:
        ...
    


class SyncParameterDecorator(BaseParameterDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        ...
    


class AsyncParameterDecorator(BaseParameterDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        ...
    


def prep_kwargs(request: t.Union[WSGIRequest, ConnexionRequest], *, request_body: t.Any, files: t.Dict[str, t.Any], arguments: t.List[str], has_kwargs: bool, sanitize: t.Callable) -> dict:
    ...

def unwrap_decorators(function: t.Callable) -> t.Callable:
    """Unwrap decorators to return the original function."""
    ...

def snake_and_shadow(name: str) -> str:
    """
    Converts the given name into Pythonic form. Firstly it converts CamelCase names to snake_case. Secondly it looks to
    see if the name matches a known built-in and if it does it appends an underscore to the name.
    :param name: The parameter name
    """
    ...

def sanitized(name: str) -> str:
    ...

def pythonic(name: str) -> str:
    ...

def get_arguments(operation: AbstractOperation, *, path_params: dict, query_params: dict, body: t.Any, files: dict, arguments: t.List[str], has_kwargs: bool, sanitize: t.Callable, content_type: str) -> t.Dict[str, t.Any]:
    """
    get arguments for handler function
    """
    ...




"""
This type stub file was generated by pyright.
"""

from typing import Any, Protocol, TypeVar

from flask import Flask, Response

A = TypeVar("A", covariant=True)
class UserMixinProtocol(Protocol[A]):
    @property
    def is_active(self) -> bool: ...

    @property
    def is_authenticated(self) -> bool: ...

    @property
    def is_anonymous(self) -> bool: ...

    def get_id(self) -> str: ...

TUserMixin = TypeVar("TUserMixin", bound=UserMixinProtocol[Any], covariant=True)

class UserLoaderCallback(Protocol[TUserMixin]):
    def __call__(self, username: str) -> TUserMixin | None:
        ...


class LoginManager:
    """This object is used to hold the settings used for logging in. Instances
    of :class:`LoginManager` are *not* bound to specific apps, so you can
    create one in the main body of your code and then bind it to your
    app in a factory function.
    """
    def __init__(self, app=..., add_context_processor=...) -> None:
        ...
    
    def setup_app(self, app, add_context_processor=...) -> None:
        """
        This method has been deprecated. Please use
        :meth:`LoginManager.init_app` instead.
        """
        ...
    
    def init_app(self, app: Flask, add_context_processor: bool=True) -> None:
        """
        Configures an application. This registers an `after_request` call, and
        attaches this `LoginManager` to it as `app.login_manager`.

        :param app: The :class:`flask.Flask` object to configure.
        :type app: :class:`flask.Flask`
        :param add_context_processor: Whether to add a context processor to
            the app that adds a `current_user` variable to the template.
            Defaults to ``True``.
        :type add_context_processor: bool
        """
        ...
    
    def unauthorized(self) -> Response:
        """
        This is called when the user is required to log in. If you register a
        callback with :meth:`LoginManager.unauthorized_handler`, then it will
        be called. Otherwise, it will take the following actions:

            - Flash :attr:`LoginManager.login_message` to the user.

            - If the app is using blueprints find the login view for
              the current blueprint using `blueprint_login_views`. If the app
              is not using blueprints or the login view for the current
              blueprint is not specified use the value of `login_view`.

            - Redirect the user to the login view. (The page they were
              attempting to access will be passed in the ``next`` query
              string variable, so you can redirect there if present instead
              of the homepage. Alternatively, it will be added to the session
              as ``next`` if USE_SESSION_FOR_NEXT is set.)

        If :attr:`LoginManager.login_view` is not defined, then it will simply
        raise a HTTP 401 (Unauthorized) error instead.

        This should be returned from a view or before/after_request function,
        otherwise the redirect will have no effect.
        """
        ...
    
    def user_loader(self, callback: UserLoaderCallback[A]) -> UserLoaderCallback[A]:
        """
        This sets the callback for reloading a user from the session. The
        function you set should take a user ID (a ``str``) and return a
        user object, or ``None`` if the user does not exist.

        :param callback: The callback for retrieving a user object.
        :type callback: callable
        """
        ...
    
    @property
    def user_callback(self) -> UserLoaderCallback[A]:
        """Gets the user_loader callback set by user_loader decorator."""
        ...
    
    def request_loader(self, callback): # -> None:
        """
        This sets the callback for loading a user from a Flask request.
        The function you set should take Flask request object and
        return a user object, or `None` if the user does not exist.

        :param callback: The callback for retrieving a user object.
        :type callback: callable
        """
        ...
    
    @property
    def request_callback(self): # -> None:
        """Gets the request_loader callback set by request_loader decorator."""
        ...
    
    def unauthorized_handler(self, callback):
        """
        This will set the callback for the `unauthorized` method, which among
        other things is used by `login_required`. It takes no arguments, and
        should return a response to be sent to the user instead of their
        normal view.

        :param callback: The callback for unauthorized users.
        :type callback: callable
        """
        ...
    
    def needs_refresh_handler(self, callback):
        """
        This will set the callback for the `needs_refresh` method, which among
        other things is used by `fresh_login_required`. It takes no arguments,
        and should return a response to be sent to the user instead of their
        normal view.

        :param callback: The callback for unauthorized users.
        :type callback: callable
        """
        ...
    
    def needs_refresh(self): # -> Response:
        """
        This is called when the user is logged in, but they need to be
        reauthenticated because their session is stale. If you register a
        callback with `needs_refresh_handler`, then it will be called.
        Otherwise, it will take the following actions:

            - Flash :attr:`LoginManager.needs_refresh_message` to the user.

            - Redirect the user to :attr:`LoginManager.refresh_view`. (The page
              they were attempting to access will be passed in the ``next``
              query string variable, so you can redirect there if present
              instead of the homepage.)

        If :attr:`LoginManager.refresh_view` is not defined, then it will
        simply raise a HTTP 401 (Unauthorized) error instead.

        This should be returned from a view or before/after_request function,
        otherwise the redirect will have no effect.
        """
        ...
    
    def header_loader(self, callback):
        """
        This function has been deprecated. Please use
        :meth:`LoginManager.request_loader` instead.

        This sets the callback for loading a user from a header value.
        The function you set should take an authentication token and
        return a user object, or `None` if the user does not exist.

        :param callback: The callback for retrieving a user object.
        :type callback: callable
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

import typing as t
from jsonschema.exceptions import ValidationError
from starlette.exceptions import HTTPException

"""
This module defines Exception classes used by Connexion to generate a proper response.
"""
class ConnexionException(Exception):
    """Base class for any exception thrown by the Connexion framework."""
    ...


class ResolverError(LookupError, ConnexionException):
    """Error raised at startup when the resolver cannot find a view function for an endpoint in
    your specification, and no ``resolver_error`` is configured."""
    ...


class InvalidSpecification(ValidationError, ConnexionException):
    """Error raised at startup when the provided specification cannot be validated."""
    ...


class MissingMiddleware(ConnexionException):
    """Error raised when you're leveraging behavior that depends on a specific middleware,
    and that middleware is not part of your middleware stack."""
    ...


class ProblemException(HTTPException, ConnexionException):
    """
    This exception holds arguments that are going to be passed to the
    `connexion.problem` function to generate a proper response.
    """
    def __init__(self, *, status=..., title=..., detail=..., type=..., instance=..., headers=..., ext=...) -> None:
        ...
    
    def to_problem(self): # -> ConnexionResponse:
        ...
    


class ClientProblem(ProblemException):
    """Base exception for any 4XX error. Returns 400 by default, however
    :class:`BadRequestProblem` should be preferred for 400 errors."""
    def __init__(self, status: int = ..., title: str = ..., *, detail: str = ...) -> None:
        ...
    


class BadRequestProblem(ClientProblem):
    """Problem class for 400 Bad Request errors."""
    def __init__(self, detail=...) -> None:
        ...
    


class ExtraParameterProblem(BadRequestProblem):
    """Problem class for 400 Bad Request errors raised when extra query or form parameters are
    detected and ``strict_validation`` is enabled."""
    def __init__(self, *, param_type: str, extra_params: t.Iterable[str]) -> None:
        ...
    


class TypeValidationError(BadRequestProblem):
    """Problem class for 400 Bad Request errors raised when path, query or form parameters with
    an incorrect type are detected."""
    def __init__(self, schema_type: str, parameter_type: str, parameter_name: str) -> None:
        ...
    


class Unauthorized(ClientProblem):
    """Problem class for 401 Unauthorized errors."""
    description = ...
    def __init__(self, detail: str = ...) -> None:
        ...
    


class OAuthProblem(Unauthorized):
    """Problem class for 401 Unauthorized errors raised when there is an issue with the received
    OAuth headers."""
    ...


class OAuthResponseProblem(OAuthProblem):
    """Problem class for 401 Unauthorized errors raised when improper OAuth credentials are
    retrieved from your OAuth server."""
    ...


class Forbidden(HTTPException):
    """Problem class for 403 Unauthorized errors."""
    def __init__(self, detail: t.Optional[str] = ...) -> None:
        ...
    


class OAuthScopeProblem(Forbidden):
    """Problem class for 403 Unauthorized errors raised because of OAuth scope validation errors."""
    def __init__(self, token_scopes: list, required_scopes: list) -> None:
        ...
    


class UnsupportedMediaTypeProblem(ClientProblem):
    """Problem class for 415 Unsupported Media Type errors which are raised when Connexion
    receives a request with an unsupported media type header."""
    def __init__(self, detail: t.Optional[str] = ...) -> None:
        ...
    


class ServerError(ProblemException):
    """Base exception for any 5XX error. Returns 500 by default, however
    :class:`InternalServerError` should be preferred for 500 errors."""
    def __init__(self, status: int = ..., title: t.Optional[str] = ..., *, detail: t.Optional[str] = ...) -> None:
        ...
    


class InternalServerError(ServerError):
    """Problem class for 500 Internal Server errors."""
    def __init__(self, detail: t.Optional[str] = ...) -> None:
        ...
    


class NonConformingResponse(InternalServerError):
    """Problem class for 500 Internal Server errors raised because of a returned response not
    matching the specification if response validation is enabled."""
    def __init__(self, detail: t.Optional[str] = ...) -> None:
        ...
    


class NonConformingResponseBody(NonConformingResponse):
    """Problem class for 500 Internal Server errors raised because of a returned response body not
    matching the specification if response validation is enabled."""
    def __init__(self, detail: t.Optional[str] = ...) -> None:
        ...
    


class NonConformingResponseHeaders(NonConformingResponse):
    """Problem class for 500 Internal Server errors raised because of a returned response headers
    not matching the specification if response validation is enabled."""
    def __init__(self, detail: t.Optional[str] = ...) -> None:
        ...
    


class ResolverProblem(ServerError):
    """Problem class for 501 Not Implemented errors raised when the resolver cannot find a view
    function to handle the incoming request."""
    def __init__(self, status: int = ..., *, detail: t.Optional[str] = ...) -> None:
        ...
    





from dataclasses import field
from os import environ
from typing import Literal

from flask.config import Config as FlaskAppConfig
from Ligare.programming.config import AbstractConfig
from pydantic import BaseModel
from typing_extensions import override


class LoggingConfig(BaseModel):
    log_level: str = "INFO"
    format: Literal["plaintext", "JSON"] = "JSON"


class WebSecurityCorsConfig(BaseModel):
    origins: list[str] | None = None
    allow_credentials: bool = False
    allow_methods: list[
        Literal[
            "GET",
            "POST",
            "PATCH",
            "PUT",
            "DELETE",
            "OPTIONS",
            "HEAD",
            "CONNECT",
            "TRACE",
        ]
    ] = field(default_factory=lambda: ["GET", "POST", "OPTIONS"])
    allow_headers: list[str | Literal["*"]] = field(default_factory=lambda: ["*"])


class WebSecurityConfig(BaseModel):
    cors: WebSecurityCorsConfig = WebSecurityCorsConfig()
    csp: str | None = None


class WebConfig(BaseModel):
    security: WebSecurityConfig = WebSecurityConfig()


class FlaskOpenApiConfig(BaseModel):
    spec_path: str | None = None
    validate_responses: bool = False
    use_swagger: bool = True
    swagger_url: str | None = None


class FlaskSessionCookieConfig(BaseModel):
    # FIXME this needs to be handled much more securely.
    # FIXME This is not done at the moment solely because we are not making
    # FIXME active use of sessions, but this should not be forgotten!
    secret_key: str | None = None
    name: str = "session"
    httponly: bool = True
    secure: bool = True
    samesite: str = "none"

    def _prepare_env_for_flask(self):
        if not self.secret_key:
            raise Exception("`flask.session.cookie.secret_key` must be set in config.")

        environ.update({
            "SECRET_KEY": self.secret_key,
            "SESSION_COOKIE_NAME": self.name,
            "SESSION_COOKIE_HTTPONLY": str(1 if self.httponly else 0),
            "SESSION_COOKIE_SECURE": str(1 if self.secure else 0),
            "SESSION_COOKIE_SAMESITE": self.samesite,
        })

    def _update_flask_config(self, flask_app_config: FlaskAppConfig):
        class ConfigObject:
            SECRET_KEY = self.secret_key
            SESSION_COOKIE_NAME = self.name
            SESSION_COOKIE_HTTPONLY = self.httponly
            SESSION_COOKIE_SECURE = self.secure
            SESSION_COOKIE_SAMESITE = self.samesite

        flask_app_config.from_object(ConfigObject)


class FlaskSessionConfig(BaseModel):
    cookie: FlaskSessionCookieConfig
    permanent: bool = True
    lifetime: int | None = 86400
    refresh_each_request: bool = True

    def _prepare_env_for_flask(self):
        environ.update({
            "PERMANENT_SESSION": str(1 if self.permanent else 0),
            "PERMANENT_SESSION_LIFETIME": (str(self.lifetime) if self.lifetime else ""),
            "SESSION_REFRESH_EACH_REQUEST": str(1 if self.refresh_each_request else 0),
        })
        self.cookie._prepare_env_for_flask()  # pyright: ignore[reportPrivateUsage]

    def _update_flask_config(self, flask_app_config: FlaskAppConfig):
        self.cookie._update_flask_config(  # pyright: ignore[reportPrivateUsage]
            flask_app_config
        )

        class ConfigObject:
            PERMANENT_SESSION = self.permanent
            PERMANENT_SESSION_LIFETIME = self.lifetime
            SESSION_REFRESH_EACH_REQUEST = self.refresh_each_request

        flask_app_config.from_object(ConfigObject)


class FlaskConfig(BaseModel):
    app_name: str = "app"
    env: str = "Development"
    host: str = "localhost"
    port: str = "5000"
    openapi: FlaskOpenApiConfig | None = None
    session: FlaskSessionConfig | None = None

    def _prepare_env_for_flask(self):
        environ.update(
            FLASK_APP=self.app_name,
            ENV=self.env,
            FLASK_RUN_PORT=self.port,
            FLASK_RUN_HOST=self.host,
        )
        if self.session:
            self.session._prepare_env_for_flask()  # pyright: ignore[reportPrivateUsage]

    def _update_flask_config(self, flask_app_config: FlaskAppConfig):
        if self.session:
            self.session._update_flask_config(  # pyright: ignore[reportPrivateUsage]
                flask_app_config
            )

        class ConfigObject:
            FLASK_APP = self.app_name
            ENV = self.env
            SERVER_NAME = f"{self.host}:{self.port}"
            FLASK_RUN_PORT = self.port
            FLASK_RUN_HOST = self.host
            TESTING = self.env == "Testing"

        flask_app_config.from_object(ConfigObject)


class Config(BaseModel, AbstractConfig):
    logging: LoggingConfig = LoggingConfig()
    web: WebConfig = WebConfig()
    flask: FlaskConfig | None = None

    def prepare_env_for_flask(self):
        if self.flask:
            self.flask._prepare_env_for_flask()  # pyright: ignore[reportPrivateUsage]

    def update_flask_config(self, flask_app_config: FlaskAppConfig):
        if self.flask:
            self.flask._update_flask_config(  # pyright: ignore[reportPrivateUsage]
                flask_app_config
            )

    @override
    def post_load(self) -> None:
        self.prepare_env_for_flask()



"""
This type stub file was generated by pyright.
"""

from connexion.operations.abstract import AbstractOperation

"""
This module defines a Swagger2Operation class, a Connexion operation specific for Swagger 2 specs.
"""
logger = ...
COLLECTION_FORMAT_MAPPING = ...
class Swagger2Operation(AbstractOperation):
    """
    Exposes a Swagger 2.0 operation under the AbstractOperation interface.
    The primary purpose of this class is to provide the `function()` method
    to the API. A Swagger2Operation is plugged into the API with the provided
    (path, method) pair. It resolves the handler function for this operation
    with the provided resolver, and wraps the handler function with multiple
    decorators that provide security, validation, serialization,
    and deserialization.
    """
    def __init__(self, method, path, operation, resolver, app_produces, app_consumes, path_parameters=..., app_security=..., security_schemes=..., definitions=..., randomize_endpoint=..., uri_parser_class=...) -> None:
        """
        :param method: HTTP method
        :type method: str
        :param path: relative path to this operation
        :type path: str
        :param operation: swagger operation object
        :type operation: dict
        :param resolver: Callable that maps operationID to a function
        :type resolver: resolver.Resolver
        :param app_produces: list of content types the application can return by default
        :type app_produces: list
        :param app_consumes: list of content types the application consumes by default
        :type app_consumes: list
        :param path_parameters: Parameters defined in the path level
        :type path_parameters: list
        :param app_security: list of security rules the application uses by default
        :type app_security: list
        :param security_schemes: `Security Definitions Object
            <https://github.com/swagger-api/swagger-spec/blob/master/versions/2.0.md#security-definitions-object>`_
        :type security_schemes: dict
        :param definitions: `Definitions Object
            <https://github.com/swagger-api/swagger-spec/blob/master/versions/2.0.md#definitionsObject>`_
        :type definitions: dict
        :param randomize_endpoint: number of random characters to append to operation name
        :type randomize_endpoint: integer
        :param uri_parser_class: class to use for uri parsing
        :type uri_parser_class: AbstractURIParser
        """
        ...
    
    @classmethod
    def from_spec(cls, spec, *args, path, method, resolver, **kwargs): # -> Self:
        ...
    
    @property
    def request_body(self) -> dict:
        ...
    
    @property
    def parameters(self):
        ...
    
    @property
    def consumes(self):
        ...
    
    @property
    def produces(self):
        ...
    
    def get_path_parameter_types(self): # -> dict[Any, Any]:
        ...
    
    def with_definitions(self, schema):
        ...
    
    def response_schema(self, status_code=..., content_type=...):
        ...
    
    def example_response(self, status_code=..., *args, **kwargs): # -> tuple[Any, int] | tuple[Any | dict[Any, Any] | list[Any], int] | tuple[None, int]:
        """
        Returns example response from spec
        """
        ...
    
    def body_name(self, content_type: str = ...) -> str:
        ...
    
    def body_schema(self, content_type: str = ...) -> dict:
        """
        The body schema definition for this operation.
        """
        ...
    
    def body_definition(self, content_type: str = ...) -> dict:
        """
        The body complete definition for this operation.

        **There can be one "body" parameter at most.**
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

import typing as t
import starlette.convertors
from starlette.types import Receive, Scope
from connexion.frameworks.abstract import Framework
from connexion.lifecycle import ConnexionRequest
from connexion.uri_parsing import AbstractURIParser

class Starlette(Framework):
    @staticmethod
    def is_framework_response(response: t.Any) -> bool:
        ...
    
    @classmethod
    def connexion_to_framework_response(cls, response): # -> JSONResponse | Response:
        ...
    
    @classmethod
    def build_response(cls, data: t.Any, *, content_type: str = ..., headers: dict = ..., status_code: int = ...): # -> JSONResponse | Response:
        ...
    
    @staticmethod
    def get_request(*, scope: Scope, receive: Receive, uri_parser: AbstractURIParser, **kwargs) -> ConnexionRequest:
        ...
    


PATH_PARAMETER = ...
PATH_PARAMETER_CONVERTERS = ...
def convert_path_parameter(match, types): # -> str:
    ...

def starlettify_path(swagger_path, types=...): # -> str:
    """
    Convert swagger path templates to flask path templates

    :type swagger_path: str
    :type types: dict
    :rtype: str

    >>> starlettify_path('/foo-bar/{my-param}')
    '/foo-bar/{my_param}'

    >>> starlettify_path('/foo/{someint}', {'someint': 'int'})
    '/foo/{someint:int}'
    """
    ...

class FloatConverter(starlette.convertors.FloatConvertor):
    """Starlette converter for OpenAPI number type"""
    regex = ...


class IntegerConverter(starlette.convertors.IntegerConvertor):
    """Starlette converter for OpenAPI integer type"""
    regex = ...





from json.decoder import JSONDecoder
from typing import Callable

import pytest
from flask import Response, abort
from Ligare.web.testing.create_app import CreateFlaskApp, FlaskClientInjector
from werkzeug.exceptions import Unauthorized


def _raise_custom_unauthorized_exception():
    response = Response("Unauthorized.", 401)
    raise Unauthorized(response.data, response)


class TestErrorHandlers(CreateFlaskApp):
    @pytest.mark.parametrize(
        "expected_status_code,expected_status_msg,failure_lambda",
        [
            (
                500,
                "Internal Server Error",
                lambda: 1 / 0,  # 1/0 to raise an exception (any exception)
            ),
            (400, "Bad Request", lambda: abort(400)),
            (401, "Unauthorized", lambda: abort(401)),
            (401, "401 UNAUTHORIZED", _raise_custom_unauthorized_exception),
        ],
    )
    def test__bind_errorhandler__calls_decorated_function_with_correct_error_when_error_occurs_during_request(
        self,
        expected_status_code: int,
        expected_status_msg: str,
        failure_lambda: Callable[[], Response],
        flask_client: FlaskClientInjector,
    ):
        _ = flask_client.client.application.route("/")(failure_lambda)

        response = flask_client.client.get("/")
        response_dict: dict[str, str | int] = JSONDecoder().decode(response.text)

        assert response_dict["error_msg"]
        assert response_dict["status_code"] == expected_status_code
        assert response_dict["status"] == expected_status_msg



"""
This type stub file was generated by pyright.
"""

from .implementation import FastAPIAppRequestInstrumentationConfigurator, FastAPIRequestAdapter, FastAPIResponseAdapter

def is_fastapi_present(): # -> bool:
    ...

if is_fastapi_present():
    ...



from .scaffolder import (  # pragma: nocover
    Operation,
    ScaffoldConfig,
    ScaffoldEndpoint,
    Scaffolder,
    ScaffoldModule,
)

__all__ = (
    "Operation",
    "ScaffoldConfig",
    "ScaffoldEndpoint",
    "Scaffolder",
    "ScaffoldModule",
)  # pragma: nocover



from Ligare.programming.collections.dict import merge
from uvicorn.workers import UvicornWorker


class LifespanUvicornWorker(UvicornWorker):
    CONFIG_KWARGS = merge(UvicornWorker.CONFIG_KWARGS, {"lifespan": "on"})


class ProxiedUvicornWorker(LifespanUvicornWorker):
    CONFIG_KWARGS = merge(
        UvicornWorker.CONFIG_KWARGS,
        {"proxy_headers": True, "forwarded_allow_ips": "*"},
    )



import pytest
from flask import Flask
from injector import Injector
from Ligare.programming.str import get_random_str
from Ligare.web.encryption import (
    _EMPTY_SECRET_KEY_MESSAGE,  # pyright:ignore[reportPrivateUsage]
)
from Ligare.web.encryption import decrypt_flask_cookie, encrypt_flask_cookie
from Ligare.web.middleware.dependency_injection import AppModule
from mock import MagicMock


def test__encrypt_flask_cookie__raises_exception_with_empty_secret_key():
    with pytest.raises(Exception, match=rf"^{_EMPTY_SECRET_KEY_MESSAGE}$"):
        _ = encrypt_flask_cookie("", {"foo": "bar"})


def test__decrypt_flask_cookie__raises_exception_with_empty_secret_key():
    with pytest.raises(Exception, match=rf"^{_EMPTY_SECRET_KEY_MESSAGE}$"):
        _ = decrypt_flask_cookie("", "")


def test__encrypt_flask_cookie__returns_serialized_dictionary():
    secret_key = get_random_str(k=26)
    cookie_value = get_random_str(k=26)
    encrypted_cookie = encrypt_flask_cookie(
        secret_key,
        {
            test__encrypt_flask_cookie__returns_serialized_dictionary.__name__: cookie_value
        },
    )

    assert encrypted_cookie
    assert isinstance(encrypted_cookie, str)


def test__decrypt_flask_cookie__returns_correct_cookie_value():
    secret_key = get_random_str(k=26)
    cookie_value = get_random_str(k=26)
    encrypted_cookie = encrypt_flask_cookie(
        secret_key,
        {
            test__decrypt_flask_cookie__returns_correct_cookie_value.__name__: cookie_value
        },
    )
    decrypted_cookie = decrypt_flask_cookie(secret_key, encrypted_cookie)

    assert isinstance(decrypted_cookie, dict)
    assert (
        decrypted_cookie[
            test__decrypt_flask_cookie__returns_correct_cookie_value.__name__
        ]
        == cookie_value
    )


def test__AppModule__binds_extra_dependencies():
    flask_mock = MagicMock(spec=Flask)
    flask_mock.name = f"{test__AppModule__binds_extra_dependencies.__name__}-app_name"
    flask_mock.config = {}
    flask_mock.configure_mock()
    extra_dependency_mock = MagicMock()

    class TestDependencyType: ...

    app_module = AppModule(flask_mock, (TestDependencyType, extra_dependency_mock))

    injector = Injector(app_module)
    resolved_type_instance = injector.get(TestDependencyType)

    assert id(resolved_type_instance) == id(extra_dependency_mock)



"""
This type stub file was generated by pyright.
"""

import typing as t
from connexion.uri_parsing import AbstractURIParser
from connexion.validators import AbstractRequestBodyValidator

logger = ...
class FormDataValidator(AbstractRequestBodyValidator):
    """Request body validator for form content types."""
    def __init__(self, *, schema: dict, required=..., nullable=..., encoding: str, strict_validation: bool, uri_parser: t.Optional[AbstractURIParser] = ...) -> None:
        ...
    


class MultiPartFormDataValidator(FormDataValidator):
    ...





"""
This type stub file was generated by pyright.
"""

from .abstract import AbstractApp

"""
This module defines Connexion applications. A Connexion App wraps a specific framework application
and exposes a standardized interface for users to create and configure their Connexion application.
"""
__all__ = ("AbstractApp",)



CORRELATION_ID_HEADER = "X-Correlation-Id"
REQUEST_COOKIE_HEADER = "Cookie"
RESPONSE_COOKIE_HEADER = "Set-Cookie"
CORS_ACCESS_CONTROL_ALLOW_ORIGIN_HEADER = "Access-Control-Allow-Origin"
CORS_ACCESS_CONTROL_ALLOW_CREDENTIALS_HEADER = "Access-Control-Allow-Credentials"
CORS_ACCESS_CONTROL_ALLOW_METHODS_HEADER = "Access-Control-Allow-Methods"
CONTENT_SECURITY_POLICY_HEADER = "Content-Security-Policy"
ORIGIN_HEADER = "Origin"
HOST_HEADER = "Host"

INCOMING_REQUEST_MESSAGE = "Incoming request:\n\
    %s %s\n\
    Host: %s\n\
    Remote address: %s\n\
    Remote user: %s"

OUTGOING_RESPONSE_MESSAGE = f"Outgoing response:\n\
   Status code: %s\n\
   Status: %s"



"""
This type stub file was generated by pyright.
"""

from jsonschema import Draft4Validator
from connexion.validators import AbstractRequestBodyValidator, AbstractResponseBodyValidator

logger = ...
class JSONRequestBodyValidator(AbstractRequestBodyValidator):
    """Request body validator for json content types."""
    def __init__(self, *, schema: dict, required=..., nullable=..., encoding: str, strict_validation: bool, **kwargs) -> None:
        ...
    


class DefaultsJSONRequestBodyValidator(JSONRequestBodyValidator):
    """Request body validator for json content types which fills in default values. This Validator
    intercepts the body, makes changes to it, and replays it for the next ASGI application."""
    MUTABLE_VALIDATION = ...
    @staticmethod
    def extend_with_set_default(validator_class):
        ...
    


class JSONResponseBodyValidator(AbstractResponseBodyValidator):
    """Response body validator for json content types."""
    @property
    def validator(self) -> Draft4Validator:
        ...
    


class TextResponseBodyValidator(JSONResponseBodyValidator):
    ...


